// DO-NOT-REMOVE begin-copyright-block 
//
// Redistributions of any form whatsoever must retain and/or include the
// following acknowledgment, notices and disclaimer:
//
// This product includes software developed by Carnegie Mellon University.
//
// Copyright 2010 by Mohammad Alisafaee, Eric Chung, Michael Ferdman, Brian Gold,
// Jangwoo Kim, Pejman Lotfi-Kamran, Onur Kocberber, Ippokratis Pandis, 
// Djordje Jevdjic, Jared Smolens, Stephen Somogyi, Evangelos Vlachos, 
// Stavros Volos, Jason Zebchuk, Babak Falsafi, Nikos Hardavellas and Tom Wenisch 
// for the SimFlex Project, Computer Architecture Lab at Carnegie Mellon,
// Carnegie Mellon University.
//
// For more information, see the SimFlex project website at:
//   http://www.ece.cmu.edu/~simflex
//
// You may not use the name "Carnegie Mellon University" or derivations
// thereof to endorse or promote products derived from this software.
//
// If you modify the software you must place a notice on or within any
// modified version provided or made available to any third party stating
// that you have modified the software.  The notice shall include at least
// your name, address, phone number, email address and the date and purpose
// of the modification.
//
// THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
// EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
// THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
// TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
// BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
// SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
// ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
// CONTRACT, TORT OR OTHERWISE).
//
// DO-NOT-REMOVE end-copyright-block   


#ifndef FLEXUS_SLICES__PERFECTPLACEMENTSLICE_HPP_INCLUDED
#define FLEXUS_SLICES__PERFECTPLACEMENTSLICE_HPP_INCLUDED

#ifdef FLEXUS_PerfectPlacementSlice_TYPE_PROVIDED
#error "Only one component may provide the Flexus::SharedTypes::PerfectPlacementSlice data type"
#endif
#define FLEXUS_PerfectPlacementSlice_TYPE_PROVIDED

#include <core/boost_extensions/intrusive_ptr.hpp>

#include <boost/optional.hpp>
#include <boost/none.hpp>

#include <core/types.hpp>
#include <core/exception.hpp>

#include <components/Common/Slices/MemoryMessage.hpp>

namespace Flexus {
namespace SharedTypes {

  using namespace Flexus::Core;
  using boost::intrusive_ptr;

  static MemoryMessage thePerfPlcDummyMemMsg(MemoryMessage::LoadReq);

  struct PerfectPlacementSlice : public boost::counted_base/*, public FastAlloc */
  {
    typedef PhysicalMemoryAddress MemoryAddress;

    // enumerated message type
    enum PerfectPlacementSliceType {
      ProcessMsg,
      MakeBlockWritable
    };

    explicit PerfectPlacementSlice(PerfectPlacementSliceType aType)
      : theType(aType)
      , theAddress(0)
      , theMemoryMessage(thePerfPlcDummyMemMsg)
    {}
    explicit PerfectPlacementSlice(PerfectPlacementSliceType aType, MemoryAddress anAddress)
      : theType(aType)
      , theAddress(anAddress)
      , theMemoryMessage(thePerfPlcDummyMemMsg)
    {}
    explicit PerfectPlacementSlice(PerfectPlacementSliceType aType, MemoryAddress anAddress, MemoryMessage & aMemMsg)
      : theType(aType)
      , theAddress(anAddress)
      , theMemoryMessage(aMemMsg)
    {}

    static intrusive_ptr<PerfectPlacementSlice> newMakeBlockWritable(MemoryAddress anAddress) {
      intrusive_ptr<PerfectPlacementSlice> slice = new PerfectPlacementSlice(MakeBlockWritable, anAddress);
      return slice;
    }
    static intrusive_ptr<PerfectPlacementSlice> newProcessMsg(MemoryMessage & aMemMsg) {
      intrusive_ptr<PerfectPlacementSlice> slice = new PerfectPlacementSlice(ProcessMsg, aMemMsg.address(), aMemMsg);
      return slice;
    }

    const PerfectPlacementSliceType type() const {
      return theType;
    }
    const MemoryAddress address() const {
      return theAddress;
    }

    PerfectPlacementSliceType & type() {
      return theType;
    }
    MemoryAddress & address() {
      return theAddress;
    }
    MemoryMessage & memMsg() {
      return theMemoryMessage;
    }

  private:
    PerfectPlacementSliceType theType;
    MemoryAddress theAddress;
    MemoryMessage & theMemoryMessage;
  };

  std::ostream & operator << (std::ostream & s, PerfectPlacementSlice const & aPerfPlcSlice);

} //End SharedTypes
} //End Flexus

#endif //FLEXUS_SLICES__PERFECTPLACEMENTSLICE_HPP_INCLUDED
