// DO-NOT-REMOVE begin-copyright-block 
//
// Redistributions of any form whatsoever must retain and/or include the
// following acknowledgment, notices and disclaimer:
//
// This product includes software developed by Carnegie Mellon University.
//
// Copyright 2010 by Mohammad Alisafaee, Eric Chung, Michael Ferdman, Brian Gold,
// Jangwoo Kim, Pejman Lotfi-Kamran, Onur Kocberber, Ippokratis Pandis, 
// Djordje Jevdjic, Jared Smolens, Stephen Somogyi, Evangelos Vlachos, 
// Stavros Volos, Jason Zebchuk, Babak Falsafi, Nikos Hardavellas and Tom Wenisch 
// for the SimFlex Project, Computer Architecture Lab at Carnegie Mellon,
// Carnegie Mellon University.
//
// For more information, see the SimFlex project website at:
//   http://www.ece.cmu.edu/~simflex
//
// You may not use the name "Carnegie Mellon University" or derivations
// thereof to endorse or promote products derived from this software.
//
// If you modify the software you must place a notice on or within any
// modified version provided or made available to any third party stating
// that you have modified the software.  The notice shall include at least
// your name, address, phone number, email address and the date and purpose
// of the modification.
//
// THE SOFTWARE IS PROVIDED "AS-IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
// EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO ANY WARRANTY
// THAT THE SOFTWARE WILL CONFORM TO SPECIFICATIONS OR BE ERROR-FREE AND ANY
// IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
// TITLE, OR NON-INFRINGEMENT.  IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY
// BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO DIRECT, INDIRECT,
// SPECIAL OR CONSEQUENTIAL DAMAGES, ARISING OUT OF, RESULTING FROM, OR IN
// ANY WAY CONNECTED WITH THIS SOFTWARE (WHETHER OR NOT BASED UPON WARRANTY,
// CONTRACT, TORT OR OTHERWISE).
//
// DO-NOT-REMOVE end-copyright-block   

#include <core/simulator_layout.hpp>

#include <components/Common/Transports/MemoryTransport.hpp>
#include <components/Common/Transports/InstructionTransport.hpp>
#include <components/Common/Slices/ArchitecturalInstruction.hpp>

#define FLEXUS_BEGIN_COMPONENT IFetch
#include FLEXUS_BEGIN_COMPONENT_DECLARATION()

COMPONENT_PARAMETERS(
  PARAMETER( StallInstructions, bool, "Stall instructions", "stalli", true )
  PARAMETER( ICacheLineSize, uint32_t, "Instruction Cache Line's Size in bytes", "icachelinesize", 64)
);

COMPONENT_INTERFACE(
  //The FetchIn port is used to read InstructionTransports from the feeder object.  Note that
  //only the ArchitecturalInstruction slice of the InstructionTransport will be valid after
  //the instruction is read from the Feeder.
  PORT( PullInput, InstructionTransport, FetchIn)
  //IFetch sends the instructions it fetches to the FetchOut port.
  PORT( PushOutput, InstructionTransport, FetchOut)
  //IFetch sends instruction commands to the Instruction Memory via this port.
  PORT( PushOutput, MemoryTransport, IMemOut )
  // Sends snoop acknowledgements out to the cache
  PORT( PushOutput, MemoryTransport, IMemSnoopOut )
  //IFetch recieves Instruction Memory responses via this port.
  PORT( PushInput, MemoryTransport, IMemIn )
  //The FeederDrive drive interface sends a commands to the Feeder and then fetches instructions,
  //then queries the instruction memory for the associated PC.
  DRIVE(FeederDrive)
  //The PipelineDrive forwards instructions along the pipeline, after a
  //response is received from the instruction memory.
  DRIVE(PipelineDrive)

);


#include FLEXUS_END_COMPONENT_DECLARATION()
#define FLEXUS_END_COMPONENT IFetch


